/*  libadfs - library to manipulate old BBC Micro/Master ADFS images
 *  Copyright (C) 2008  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  */

#include "adfs.h"
#include "debug.h"
#include "dir.h"
#include "dirent.h"
#include "disk.h"
#include "file.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#define DIR_ENTRY_SIZE (26)
#define MAX_DIRECTORY_ENTRIES (47)

ADFS_DIR *ADFS_opendir(struct ADFS_disk *disk, const char *path) {
	ADFS_FILE *fd;

	fd = ADFS_open(disk, path, O_RDONLY, 0);
	if (fd == NULL) {
		LOG_DEBUG("ADFS_opendir(): ADFS_open() returned NULL\n");
		return NULL;
	}

	return ADFS_opendir_fd(fd);
}

ADFS_DIR *ADFS_opendir_dirent(struct ADFS_dirent *ent) {
	ADFS_FILE *fd;

	LOG_DEBUG("ADFS_opendir_dirent(): lsn=%06X parent_lsn=%06X parent_index=%d\n", ent->start_lsn, ent->parent_dir_lsn, ent->parent_dir_index);
	fd = ADFS_open_dirent(ent, O_RDONLY, 0);
	if (fd == NULL)
		return NULL;
	return ADFS_opendir_fd(fd);
}

ADFS_DIR *ADFS_opendir_fd(ADFS_FILE *fd) {
	uint8_t head_buf[5];
	uint8_t foot_buf[53];
	ADFS_DIR *new_dir;
	int i;

	if (fd == NULL) {
		errno = ENOENT;
		return NULL;
	}
	if (!(fd->dirent->attr & ADFS_ATTR_DIR)) {
		errno = ENOTDIR;
		return NULL;
	}

	if (ADFS_lseek(fd, 0, SEEK_SET) < 0)
		return NULL;
	if (ADFS_read(fd, head_buf, 5) < 5)
		return NULL;
	if (ADFS_lseek(fd, 0x4cb, SEEK_SET) < 0)
		return NULL;
	if (ADFS_read(fd, foot_buf, 53) < 53)
		return NULL;

	/* Check that file looks like a directory */
	if (foot_buf[0] != 0 || foot_buf[52] != 0
			|| 0 != memcmp(head_buf, foot_buf + 47, 5)
			|| 0 != memcmp(head_buf + 1, "Hugo", 4)) {
		errno = ENOTDIR;
		return NULL;
	}

	new_dir = malloc(sizeof(ADFS_DIR));
	if (new_dir == NULL) {
		ADFS_close(fd);
		return NULL;
	}

	new_dir->fd = fd;
	new_dir->index = -2;
	new_dir->cycle_number = head_buf[0];

	/* Read directory name */
	for (i = 0; i < 10; i++) {
		new_dir->name[i] = foot_buf[1 + i];
		if (new_dir->name[i] == 13)
			break;
	}
	new_dir->name[i] = 0;

	/* Read parent LSN */
	new_dir->fd->dirent->parent_dir_lsn = foot_buf[11] | (foot_buf[12] << 8) | (foot_buf[13] << 16);

	/* Read directory title */
	for (i = 0; i < 19; i++) {
		new_dir->title[i] = foot_buf[14 + i];
		if (new_dir->title[i] == 13)
			break;
	}
	new_dir->title[i] = 0;

	return new_dir;
}

void ADFS_closedir(ADFS_DIR *dir) {
	if (dir == NULL)
		return;
	ADFS_close(dir->fd);
	free(dir);
}

int ADFS_readdir_r(ADFS_DIR *dir, struct ADFS_dirent *entry, struct ADFS_dirent **result) {
	uint8_t ent_buf[DIR_ENTRY_SIZE];
	int i;
	*result = NULL;
	if (dir == NULL)
		return ENOTDIR;
	switch (dir->index) {
		case -2:
			copy_dirent(entry, dir->fd->dirent);
			strcpy(entry->name, ".");
			dir->index++;
			break;
		case -1:
			if (parent_dirent(entry, dir->fd->dirent) < 0)
				return 0;
			strcpy(entry->name, "..");
			dir->index++;
			break;
		default:
			ent_buf[0] = 0;
			/* Loop until a valid directory entry found */
			while ((ent_buf[0] & 0x7f) == 0) {
				if (dir->index >= MAX_DIRECTORY_ENTRIES)
					return 0;
				if (ADFS_lseek(dir->fd, 5 + dir->index * DIR_ENTRY_SIZE, SEEK_SET) < 0)
					return errno;
				if (ADFS_read(dir->fd, ent_buf, DIR_ENTRY_SIZE) < DIR_ENTRY_SIZE)
					return ENOENT;
				dir->index++;
			}
			entry->attr = 0; 
			for (i = 0; i < 10; i++) {
				if (ent_buf[i] & 0x80)
					entry->attr |= (1 << i);
				entry->name[i] = ent_buf[i] & 0x7f;
				if (entry->name[i] == '/')
					entry->name[i] = '.';
				if (entry->name[i] == 13)
					entry->name[i] = 0;
			}
			entry->name[10] = 0;
			entry->load_address = ent_buf[10] | (ent_buf[11] << 8) | (ent_buf[12] << 16) | (ent_buf[13] << 24);
			entry->exec_address = ent_buf[14] | (ent_buf[15] << 8) | (ent_buf[16] << 16) | (ent_buf[17] << 24);
			entry->length = ent_buf[18] | (ent_buf[19] << 8) | (ent_buf[20] << 16) | (ent_buf[21] << 24);
			entry->start_lsn = ent_buf[22] | (ent_buf[23] << 8) | (ent_buf[24] << 16);
			entry->cycle_number = ent_buf[25];
			entry->disk = dir->fd->dirent->disk;
			entry->parent_dir_lsn = dir->fd->dirent->start_lsn;
			entry->parent_dir_index = dir->index - 1;
			break;
	}
	*result = entry;
	return 0;
}

struct ADFS_dirent *ADFS_readdir(ADFS_DIR *dir) {
	struct ADFS_dirent *result;
	int err = ADFS_readdir_r(dir, &dir->ent_buf, &result);
	if (err != 0)
		errno = err;
	return result;
}

void ADFS_rewinddir(ADFS_DIR *dir) {
	if (dir == NULL)
		return;
	dir->index = -2;
}

static struct ADFS_dirent *dirents[MAX_DIRECTORY_ENTRIES];
static struct ADFS_dirent dirent_buf[MAX_DIRECTORY_ENTRIES];
static int num_dirents;

static int increment_dir_cycle(ADFS_DIR *dir) {
	if (dir == NULL || dir->fd == NULL) {
		errno = EBADF;
		return -1;
	}
	dir->cycle_number++;
	if ((dir->cycle_number & 0x0f) > 0x09)
		dir->cycle_number += 6;
	if ((dir->cycle_number & 0xf0) > 0x90)
		dir->cycle_number = 0;
	return 0;
}

int read_entire_directory(ADFS_DIR *dir) {
	struct ADFS_dirent *tmp;
	num_dirents = 0;
	dir->index = 0;  /* skip . and .. */
	do {
		if (ADFS_readdir_r(dir, &dirent_buf[num_dirents], &tmp) != 0)
			return -1;
		if (tmp) {
			dirents[num_dirents] = &dirent_buf[num_dirents];
			num_dirents++;
		}
	} while (tmp);
	return 0;
}

static int compar_dirent_byname(const void *a, const void *b) {
	struct ADFS_dirent *aa = *((struct ADFS_dirent * const *)a);
	struct ADFS_dirent *bb = *((struct ADFS_dirent * const *)b);
	if (aa->name[0] == 0 && bb->name[0] == 0)
		return 0;
	if (aa->name[0] == 0)
		return 1;
	if (bb->name[0] == 0)
		return -1;
	return strcasecmp(aa->name, bb->name);
}

int write_entire_directory(ADFS_DIR *dir) {
	uint8_t ent_buf[DIR_ENTRY_SIZE];
	char cycle;
	int i;
	qsort(dirents, num_dirents, sizeof(struct ADFS_dirent *), compar_dirent_byname);
	dir->fd->oflag = O_RDWR;
	cycle = dir->cycle_number;
	ADFS_lseek(dir->fd, 0, SEEK_SET);
	ADFS_write(dir->fd, &cycle, 1);
	for (i = 0; i < num_dirents; i++) {
		if (dirents[i]->name[0] == 0)
			break;
		dirents[i]->parent_dir_index = i;
		update_directory_entry(dir, dirents[i]);
	}
	memset(ent_buf, 0, DIR_ENTRY_SIZE);
	for ( ; i < MAX_DIRECTORY_ENTRIES; i++) {
		ADFS_lseek(dir->fd, 5 + i * DIR_ENTRY_SIZE, SEEK_SET);
		ADFS_write(dir->fd, ent_buf, DIR_ENTRY_SIZE);
	}
	ADFS_lseek(dir->fd, 0x4fa, SEEK_SET);
	ADFS_write(dir->fd, &cycle, 1);
	return 0;
}

int update_directory_entry(ADFS_DIR *dir, struct ADFS_dirent *ent) {
	uint8_t ent_buf[DIR_ENTRY_SIZE];
	int i;

	if (dir == NULL || ent == NULL)
		return -1;
	LOG_DEBUG("Updating index=%d dir=%06X name=%s lsn=%06X length=%08X\n", ent->parent_dir_index, dir->fd->dirent->start_lsn, ent->name, ent->start_lsn, ent->length);
	for (i = 0; i < 10; i++) {
		int eflag = 0;
		if (eflag) {
			ent_buf[i] = 0;
		} else {
			ent_buf[i] = ent->name[i];
			if (ent_buf[i] == '.')
				ent_buf[i] = '/';
			if (ent_buf[i] == 0) {
				eflag = 1;
				ent_buf[i] = 13;
			}
		}
		if (ent->attr & (1 << i))
			ent_buf[i] |= 0x80;
	}
	ent_buf[10] = ent->load_address & 0xff;
	ent_buf[11] = (ent->load_address >> 8) & 0xff;
	ent_buf[12] = (ent->load_address >> 16) & 0xff;
	ent_buf[13] = (ent->load_address >> 24) & 0xff;
	ent_buf[14] = ent->exec_address & 0xff;
	ent_buf[15] = (ent->exec_address >> 8) & 0xff;
	ent_buf[16] = (ent->exec_address >> 16) & 0xff;
	ent_buf[17] = (ent->exec_address >> 24) & 0xff;
	ent_buf[18] = ent->length & 0xff;
	ent_buf[19] = (ent->length >> 8) & 0xff;
	ent_buf[20] = (ent->length >> 16) & 0xff;
	ent_buf[21] = (ent->length >> 24) & 0xff;
	ent_buf[22] = ent->start_lsn & 0xff;
	ent_buf[23] = (ent->start_lsn >> 8) & 0xff;
	ent_buf[24] = (ent->start_lsn >> 16) & 0xff;
	ent_buf[25] = ent->cycle_number;
	dir->fd->oflag = O_RDWR;
	ADFS_lseek(dir->fd, 5 + ent->parent_dir_index * DIR_ENTRY_SIZE, SEEK_SET);
	ADFS_write(dir->fd, ent_buf, DIR_ENTRY_SIZE);
	return 0;
}

int add_to_directory(ADFS_DIR *dir, struct ADFS_dirent *ent) {
	int ret;
	read_entire_directory(dir);
	if (num_dirents >= MAX_DIRECTORY_ENTRIES) {
		errno = ENOSPC;
		return -1;
	}
	increment_dir_cycle(dir);
	ent->cycle_number = dir->cycle_number;
	dirents[num_dirents] = ent;
	num_dirents++;
	LOG_DEBUG("add_to_directory(): adding %s. before, parent_index=%d\n", ent->name, ent->parent_dir_index);
	ret = write_entire_directory(dir);
	LOG_DEBUG("add_to_directory(): adding %s.  after, parent_index=%d\n", ent->name, ent->parent_dir_index);
	return ret;
}

int remove_from_directory(ADFS_DIR *dir, int ent_index) {
	if (ent_index < 0 || ent_index >= MAX_DIRECTORY_ENTRIES) {
		LOG_DEBUG("remove_from_directory(): bad entry %d in %06X\n", ent_index, dir->fd->dirent->start_lsn);
		errno = ENOENT;
		return -1;
	}
	LOG_DEBUG("remove_from_directory(): removing entry %d from %06X\n", ent_index, dir->fd->dirent->start_lsn);
	read_entire_directory(dir);
	dirents[ent_index]->name[0] = 0;
	increment_dir_cycle(dir);
	return write_entire_directory(dir);
}

/* All mkdir does is create a file O_EXCL with special contents and
 * ADFS_ATTR_DIR set in the attributes. */
int ADFS_mkdir(struct ADFS_disk *disk, const char *path) {
	ADFS_FILE *fd;
	uint8_t head_buf[5];
	uint8_t ent_buf[DIR_ENTRY_SIZE];
	uint8_t foot_buf[53];
	int i;

	fd = ADFS_open(disk, path, O_CREAT | O_EXCL | O_WRONLY,
			ADFS_ATTR_DIR | ADFS_ATTR_LOCKED | ADFS_ATTR_READ);
	if (fd == NULL)
		return -1;
	if (ADFS_ftruncate(fd, DIRECTORY_LENGTH) < 0) {
		ADFS_close(fd);
		return -1;
	}
	LOG_DEBUG("ADFS_mkdir(): created file, lsn=%06X length=%X parent=%06X\n", fd->dirent->start_lsn, fd->dirent->length, fd->dirent->parent_dir_lsn);
	head_buf[0] = fd->dirent->cycle_number;
	memcpy(head_buf + 1, "Hugo", 4);
	ADFS_write(fd, head_buf, 5);
	memset(ent_buf, 0, DIR_ENTRY_SIZE);
	for (i = 0; i < 47; i++)
		ADFS_write(fd, ent_buf, DIR_ENTRY_SIZE);
	memset(foot_buf, 0, 53);
	for (i = 0; i < 10; i++) {
		if (fd->dirent->name[i] == 0)
			break;
		foot_buf[1 + i] = fd->dirent->name[i];
		if (foot_buf[1 + i] == '.')
			foot_buf[1 + i] = '/';
		foot_buf[14 + i] = foot_buf[1 + i];
	}
	if (i < 10) {
		foot_buf[1 + i] = 13;
	}
	foot_buf[14 + i] = 13;
	foot_buf[11] = fd->dirent->parent_dir_lsn & 0xff;
	foot_buf[12] = (fd->dirent->parent_dir_lsn >> 8) & 0xff;
	foot_buf[13] = (fd->dirent->parent_dir_lsn >> 16) & 0xff;
	memcpy(foot_buf + 47, head_buf, 5);
	ADFS_write(fd, foot_buf, 53);
	ADFS_close(fd);
	return 0;
}

int ADFS_chdir(struct ADFS_disk *disk, const char *path) {
	ADFS_DIR *dir;
	struct ADFS_dirent *ent;

	if (find_dir_and_ent(disk, path, &dir, &ent) < 0)
		return -1;
	if (dir)
		ADFS_closedir(dir);
	if (ent == NULL)
		return -1;
	if (disk->cwd_dirent)
		free(disk->cwd_dirent);
	disk->cwd_dirent = dup_dirent(ent);

	return 0;
}

int ADFS_fchdir(ADFS_FILE *fd) {
	if (fd == NULL || fd->dirent == NULL) {
		errno = EBADF;
		return -1;
	}
	if (!(fd->dirent->attr & ADFS_ATTR_DIR)) {
		errno = ENOTDIR;
		return -1;
	}
	if (fd->dirent->disk->cwd_dirent)
		free(fd->dirent->disk->cwd_dirent);
	fd->dirent->disk->cwd_dirent = dup_dirent(fd->dirent);
	return 0;
}
