/*  libadfs - library to manipulate old BBC Micro/Master ADFS images
 *  Copyright (C) 2008  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  */

#include "adfs.h"
#include "debug.h"
#include "dir.h"
#include "dirent.h"
#include "disk.h"
#include "file.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>

struct ADFS_dirent *dup_dirent(struct ADFS_dirent *ent) {
	struct ADFS_dirent *new;
	if (ent == NULL)
		return NULL;
	new = malloc(sizeof(struct ADFS_dirent));
	if (new == NULL)
		return NULL;
	memcpy(new, ent, sizeof(struct ADFS_dirent));
	return new;
}

struct ADFS_dirent *copy_dirent(struct ADFS_dirent *dst,
		struct ADFS_dirent *src) {
	return memcpy(dst, src, sizeof(struct ADFS_dirent));
}

int truncate_dirent(struct ADFS_dirent *ent, off_t length) {
	int old_size, new_size, new_start_lsn, i;
	int num_fsm_entries;
	struct fsm_entry *fsm;

	if (ent == NULL) {
		errno = EBADF;
		return -1;
	}
	if (length < 0) {
		errno = EINVAL;
		return -1;
	}
	if ((unsigned int)length == ent->length)
		return 0;
	old_size = (ent->length + SECTOR_SIZE - 1) / SECTOR_SIZE;
	new_size = (length + SECTOR_SIZE - 1) / SECTOR_SIZE;
	if (new_size == old_size) {
		//LOG_DEBUG("truncate_dirent(): size stays the same\n");
		ent->length = length;
		return 0;
	}
	new_start_lsn = ent->start_lsn;
	num_fsm_entries = ent->disk->num_fsm_entries;
	fsm = ent->disk->fsm;
	ent->disk->modified = 1;
	if (ent->length == 0) {
		for (i = 0; i < num_fsm_entries; i++) {
			if (fsm[i].size >= new_size) {
				new_start_lsn = fsm[i].lsn;
				break;
			}
		}
	}
	if (new_start_lsn <= 0) {
		LOG_DEBUG("truncate_dirent(): failed to find space to start file\n");
		errno = ENOSPC;
		return -1;
	}

	LOG_DEBUG("truncate_dirent(): '%s' old=%06X-%06X(%d), new=%06X-%06X(%d)\n", ent->name, ent->start_lsn, ent->start_lsn+old_size-1, old_size, new_start_lsn, new_start_lsn+new_size-1, new_size);
	if (new_size < old_size) {
		/* find adjacent free space and grow it downwards */
		for (i = 0; i < num_fsm_entries; i++) {
			if (fsm[i].lsn == new_start_lsn + old_size) {
				LOG_DEBUG("truncate_dirent(): growing fsm entry from %06X(%d)", fsm[i].lsn, fsm[i].size);
				fsm[i].lsn -= (old_size - new_size);
				fsm[i].size += (old_size - new_size);
				LOG_DEBUG(" to %06X(%d)\n", fsm[i].lsn, fsm[i].size);
				ent->start_lsn = new_start_lsn;
				ent->length = length;
				coalesce_fsm(ent->disk);
				ent->modified = 1;
				return 0;
			}
		}
		/* otherwise, try adding a new free space entry */
		if (num_fsm_entries < MAX_FSM_ENTRIES) {
			fsm[num_fsm_entries].lsn = new_start_lsn + new_size;
			fsm[num_fsm_entries].size = old_size - new_size;
			LOG_DEBUG("truncate_dirent(): adding new fsm entry %06X(%d)\n", fsm[num_fsm_entries].lsn, fsm[num_fsm_entries].size);
			ent->disk->num_fsm_entries++;
			ent->start_lsn = new_start_lsn;
			ent->length = length;
			coalesce_fsm(ent->disk);
			ent->modified = 1;
			return 0;
		}
		LOG_DEBUG("truncate_dirent(): no space to add new fsm entry!\n");
		/* Can't *reduce* space...  What to do? */
		errno = ENOSPC;
		return -1;
	}
	/* find adjacent free space and shrink it */
	for (i = 0; i < num_fsm_entries; i++) {
		if (fsm[i].lsn == new_start_lsn + old_size) {
			if (fsm[i].size >= (new_size - old_size)) {
				LOG_DEBUG("truncate_dirent(): shrinking fsm entry from %06X(%d)", fsm[i].lsn, fsm[i].size);
				fsm[i].lsn += (new_size - old_size);
				fsm[i].size -= (new_size - old_size);
				LOG_DEBUG(" to %06X(%d)\n", fsm[i].lsn, fsm[i].size);
				if (fsm[i].size == 0) {
					LOG_DEBUG("truncate_dirent(): deleting fsm entry %d: %06X(%d)\n", i, fsm[i].lsn, fsm[i].size);
					for ( ; i < num_fsm_entries + 1; i++) {
						fsm[i].lsn = fsm[i+1].lsn;
						fsm[i].size = fsm[i+1].size;
					}
					ent->disk->num_fsm_entries--;
				}
				ent->start_lsn = new_start_lsn;
				ent->length = length;
				/* coalesce should not be required */
				ent->modified = 1;
				return 0;
			}
		}
	}
	LOG_DEBUG("truncate_dirent(): no free space\n");
	errno = ENOSPC;
	return -1;
}

/* Construct a dirent for the parent of ent.  Opens it as a directory to find
 * the parent's name and grandparent lsn.  */

int parent_dirent(struct ADFS_dirent *into, struct ADFS_dirent *ent) {
	struct ADFS_dirent parent;
	ADFS_DIR *dir;
	parent.name[0] = 0;
	parent.attr = ADFS_ATTR_DIR | ADFS_ATTR_LOCKED | ADFS_ATTR_READ;
	parent.load_address = 0;
	parent.exec_address = 0;
	parent.length = DIRECTORY_LENGTH;
	parent.start_lsn = ent->parent_dir_lsn;
	parent.cycle_number = 0;
	parent.disk = ent->disk;
	parent.parent_dir_lsn = -1;
	parent.parent_dir_index = -1;
	parent.modified = 0;
	/* opendir will read parent lsn from the directory file */
	dir = ADFS_opendir_dirent(&parent);
	if (dir == NULL)
		return -1;
	strncpy(parent.name, dir->name, 10);
	parent.name[10] = 0;
	parent.parent_dir_lsn = dir->fd->dirent->parent_dir_lsn;
	ADFS_closedir(dir);
	copy_dirent(into, &parent);
	return 0;
}

int find_dir_and_ent(struct ADFS_disk *disk, const char *path,
		ADFS_DIR **dirp, struct ADFS_dirent **entp) {
	static struct ADFS_dirent ent;
	struct ADFS_dirent *tmp;
	ADFS_DIR *dir;
	char *pathc, *tok, *next_tok;

	if (dirp)
		*dirp = NULL;
	if (entp)
		*entp = NULL;

	if (disk == NULL) {
		errno = ENODEV;
		return -1;
	}

	if (path == NULL || *path == 0) {
		errno = ENOENT;
		return -1;
	}

	if (*path == '/') {
		copy_dirent(&ent, &disk->root_dirent);
		path++;
	} else {
		copy_dirent(&ent, disk->cwd_dirent);
	}

	pathc = strdup(path);
	tok = NULL;
	next_tok = strtok(pathc, "/");
	tmp = &ent;

	/* For each / separated token bar the last, open as a directory and
	 * search it for the next one.  Leave the last one in ent.  */

	dir = NULL;
	while (next_tok) {
		tok = next_tok;
		next_tok = strtok(NULL, "/");
		if (dir)
			ADFS_closedir(dir);
		dir = ADFS_opendir_dirent(&ent);
		if (dir == NULL) {
			free(pathc);
			return -1;
		}
		while ((tmp = ADFS_readdir(dir))) {
			if (0 == strcasecmp(tmp->name, tok))
				break;
		}
		if (tmp == NULL) {
			if (next_tok) {
				free(pathc);
				errno = ENOTDIR;
				return -1;
			}
		} else {
			copy_dirent(&ent, tmp);
		}
	}

	/* If the last token isn't found, create a dummy entry in case the
	 * calling function wants to create it.  */

	if (tmp == NULL && tok) {
		strncpy(ent.name, tok, 10);
		ent.name[10] = 0;
		ent.load_address = 0;
		ent.exec_address = 0;
		ent.length = 0;
		ent.start_lsn = 0;
		ent.disk = disk;
		if (dir)
			ent.parent_dir_lsn = dir->fd->dirent->start_lsn;
		else
			ent.parent_dir_lsn = -1;
		ent.parent_dir_index = -2;
		ent.modified = 1;
		LOG_DEBUG("find_dir_and_ent(): created dummy entry for %s parent_lsn=%06X\n", ent.name, ent.parent_dir_lsn);
	}

	if (dirp)
		*dirp = dir;
	if (entp)
		*entp = &ent;
	free(pathc);
	LOG_DEBUG("find_dir_and_ent(): returning with ent->parent_dir_index = %d\n", ent.parent_dir_index);
	return 0;
}
