/*  libadfs - library to manipulate old BBC Micro/Master ADFS images
 *  Copyright (C) 2008  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  */

#include "adfs.h"
#include "debug.h"
#include "dirent.h"
#include "disk.h"

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>

#define MAX_SECTORS (0x1fff00)
#define ROOT_START_SECTOR (2)
#define DATA_START_SECTOR (7)
#define DIRECTORY_SECTORS (5)
#define ROOT_START_SECTOR (2)

/* read_sector and write_sector use this: */
static uint8_t rw_buf[SECTOR_SIZE * 2];
/* general-purpose, big enough to hold two sectors: */
static uint8_t gbuf[SECTOR_SIZE * 2];

/* Seems when most people dd off their double-sided ADFS floppies, they don't
 * convert the image into logical sector order.  This does the conversion.  */

static int chs_to_hcs(struct ADFS_disk *disk, int lsn) {
	int tracks_per_head = disk->total_sectors >> 5;
	int sector = lsn & 15;
	int track = lsn >> 4;
	int head = (track >= tracks_per_head) ? 1 : 0;
	return (track % tracks_per_head) * 32 + head * 16 + sector;
}

/* Read sector, accounting for 'sparse' flag */
int read_sector(struct ADFS_disk *disk, int lsn, uint8_t *buf) {
	uint8_t *tmp = rw_buf;
	int remain, tries, i;
	if (disk == NULL) {
		LOG_DEBUG("read_sector(): disk == NULL\n");
		return 0;
	}
	if (disk->fd < 0) {
		LOG_DEBUG("read_sector(): no fd\n");
		return 0;
	}
	LOG_DEBUG("read_sector(): fd %d, lsn %06X", disk->fd, lsn);
	if (disk->translate)
		lsn = chs_to_hcs(disk, lsn);
	LOG_DEBUG(" -> %06X\n", lsn);
	if (lsn >= disk->total_sectors) {
		LOG_DEBUG("read_sector(): lsn(%06X) >= total_sectors(%06X)\n", lsn, disk->total_sectors);
		return 0;
	}
	if (lseek(disk->fd, lsn * SECTOR_SIZE * disk->sparse, SEEK_SET) < 0) {
		perror(NULL);
		LOG_DEBUG("read_sector(): seek failed\n");
		return 0;
	}
	remain = SECTOR_SIZE * disk->sparse;
	tries = 3;
	while (remain > 0 && tries > 0) {
		int r = read(disk->fd, tmp, remain);
		if (r >= 0) {
			remain -= r;
			tmp += r;
		}
		tries--;
	}
	if (remain != 0) {
		LOG_DEBUG("read_sector(): too many tries, remain=%d\n", remain);
		return 0;
	}
	for (i = 0; i < SECTOR_SIZE; i++) {
		buf[i] = rw_buf[i * disk->sparse];
		//LOG_DEBUG("%02x  ", buf[i]);
		//if ((i % 16) == 15)
			//LOG_DEBUG("\n");
	}
	return 1;
}

/* Write sector, accounting for 'sparse' flag */
int write_sector(struct ADFS_disk *disk, int lsn, uint8_t *buf) {
	uint8_t *tmp = rw_buf;
	int remain, tries, i;
	if (disk == NULL) {
		LOG_DEBUG("write_sector(): disk == NULL\n");
		return 0;
	}
	if (disk->fd < 0) {
		LOG_DEBUG("write_sector(): no fd\n");
		return 0;
	}
	LOG_DEBUG("write_sector(): lsn %06X", lsn);
	if (disk->translate)
		lsn = chs_to_hcs(disk, lsn);
	LOG_DEBUG(" -> %06X\n", lsn);
	if (lsn >= disk->total_sectors) {
		LOG_DEBUG("write_sector(): lsn(%06X) >= total_sectors(%06X)\n", lsn, disk->total_sectors);
		return 0;
	}
	if (lseek(disk->fd, lsn * SECTOR_SIZE * disk->sparse, SEEK_SET) < 0) {
		LOG_DEBUG("write_sector(): seek failed\n");
		return 0;
	}
	for (i = 0; i < SECTOR_SIZE; i++) {
		rw_buf[i * disk->sparse] = buf[i];
		if (disk->sparse == 2)
			rw_buf[(i * 2)+1] = 0;
	}
	remain = SECTOR_SIZE * disk->sparse;
	tries = 3;
	while (remain > 0 && tries > 0) {
		int r = write(disk->fd, tmp, remain);
		if (r >= 0) {
			remain -= r;
			tmp += r;
		}
		tries--;
	}
	if (remain != 0) {
		LOG_DEBUG("write_sector(): too many tries\n");
		return 0;
	}
	return 1;
}

static int compute_checksum(uint8_t *sector) {
	int sum = 255, carry = 0, i;
	for (i = (SECTOR_SIZE - 2); i >= 0; i--) {
		sum += sector[i] + carry;
		carry = sum >> 8;
		sum &= 0xff;
	}
	return sum;
}

struct ADFS_disk *ADFS_open_disk(const char *filename, int oflag) {
	struct ADFS_disk *new;
	int i;
	int free_sectors = 0;
	int translate = 0;

	new = malloc(sizeof(struct ADFS_disk));
	if (new == NULL) {
		LOG_DEBUG("ADFS_open_disk(): malloc() failed\n");
		return NULL;
	}
	new->fd = -1;
	new->modified = 0;
	new->fd = open(filename, oflag);
	if (new->fd < 0) {
		LOG_DEBUG("ADFS_open_disk(): open() failed\n");
		ADFS_close_disk(new);
		return NULL;
	}

	/* To determine if this image is 'sparse', read the root directory
	 * in both modes and see which one the identifier is recognisable in */
	new->sparse = 1;
	new->translate = 0;
	/* temporarily so read_sector won't error out: */
	new->total_sectors = ROOT_START_SECTOR + 1;
	memset(gbuf, 0, 5);
	if (!read_sector(new, ROOT_START_SECTOR, gbuf)) {
		LOG_DEBUG("ADFS_open_disk(): Failed to read sector %d\n", ROOT_START_SECTOR);
		ADFS_close_disk(new);
		return NULL;
	}
	if (memcmp(&gbuf[1], "Hugo", 4) != 0) {
		new->sparse = 2;
		if (!read_sector(new, ROOT_START_SECTOR, gbuf)
				|| 0 != memcmp(&gbuf[1], "Hugo", 4)) {
			LOG_DEBUG("ADFS_open_disk(): Identifier not found in root directory\n");
			ADFS_close_disk(new);
			return NULL;
		}
		LOG_DEBUG("ADFS_open_disk(): Disk image is sparse!\n");
	}
	if (!read_sector(new, 0, gbuf)) {
		LOG_DEBUG("ADFS_open_disk(): read_sector(0) failed\n");
		ADFS_close_disk(new);
		return NULL;
	}
	if (compute_checksum(gbuf) != gbuf[255]) {
		LOG_DEBUG("ADFS_open_disk(): checksum in sector 0: computed=%02X, actual=%02X\n", compute_checksum(gbuf), gbuf[255]);
		ADFS_close_disk(new);
		return NULL;
	}
	if (!read_sector(new, 1, gbuf + SECTOR_SIZE)) {
		LOG_DEBUG("ADFS_open_disk(): read_sector(1) failed\n");
		ADFS_close_disk(new);
		return NULL;
	}
	if (compute_checksum(gbuf + SECTOR_SIZE) != gbuf[SECTOR_SIZE+255]) {
		LOG_DEBUG("ADFS_open_disk(): checksum in sector 1: computed=%02X, actual=%02X\n", compute_checksum(gbuf + SECTOR_SIZE), gbuf[SECTOR_SIZE+255]);
		ADFS_close_disk(new);
		return NULL;
	}

	new->total_sectors = gbuf[0xfc] | (gbuf[0xfd]<<8) | (gbuf[0xfe]<<16);
	//if (translate < 0) {
		/* < 0 = autodetect: translate only for sectors = &A00 (L) */
		if (new->total_sectors == 0xa00)
			translate = 1;
		else
			translate = 0;
	//}
	new->translate = translate;
	new->max_fsm_entries = MAX_FSM_ENTRIES;
	new->num_fsm_entries = gbuf[0x1fe] / 3;
	if (new->num_fsm_entries > new->max_fsm_entries)
		new->num_fsm_entries = new->max_fsm_entries;
	LOG_DEBUG("num_fsm_entries = %d\n", new->num_fsm_entries);
	new->disk_id = gbuf[0x1fb] | (gbuf[0x1fc]<<8);
	new->boot_option = gbuf[0x1fd];

	for (i = 0; i < new->num_fsm_entries; i++) {
		new->fsm[i].lsn = gbuf[i*3] | (gbuf[i*3+1]<<8) | (gbuf[i*3+2]<<16);
		new->fsm[i].size = gbuf[i*3+256] | (gbuf[i*3+257]<<8) | (gbuf[i*3+258]<<16);
		free_sectors += new->fsm[i].size;
		//LOG_DEBUG("FSM @ %d length = %d\n", new->fsm[i].sector, new->fsm[i].length);
	}
	LOG_DEBUG("ADFS_open_disk(): free space = %d\n", free_sectors);

	new->root_dirent.disk = new;
	new->root_dirent.start_lsn = ROOT_START_SECTOR;
	strcpy(new->root_dirent.name, "$");
	new->root_dirent.attr = ADFS_ATTR_DIR | ADFS_ATTR_LOCKED | ADFS_ATTR_READ;
	new->root_dirent.load_address = 0;
	new->root_dirent.exec_address = 0;
	new->root_dirent.length = DIRECTORY_SECTORS * SECTOR_SIZE;
	new->root_dirent.parent_dir_lsn = ROOT_START_SECTOR;
	new->root_dirent.parent_dir_index = -1;

	new->cwd_dirent = dup_dirent(&new->root_dirent);

	return new;
}

void ADFS_close_disk(struct ADFS_disk *disk) {
	if (disk == NULL) return;
	if (disk->modified) {
		LOG_DEBUG("ADFS_close_disk(): Disk modified - rewriting FSM\n");
		rewrite_fsm(disk);
	}
	if (disk->fd >= 0)
		close(disk->fd);
	free(disk);
}

int ADFS_get_boot_option(struct ADFS_disk *disk) {
	return disk->boot_option & 3;
}

void ADFS_set_boot_option(struct ADFS_disk *disk, int option) {
	disk->boot_option = option & 3;
	disk->modified = 1;
}

/* Disk Free Space Map (FSM) manipulation */

static int compar_fsm(const void *a, const void *b) {
	const struct fsm_entry *aa = a, *bb = b;
	if (aa->lsn < bb->lsn)
		return -1;
	if (aa->lsn > bb->lsn)
		return 1;
	return 0;
}

void coalesce_fsm(struct ADFS_disk *disk) {
	int i, cur, next_sector, num_fsm_entries;
	if (disk == NULL)
		return;
	if (disk->num_fsm_entries <= 1)
		return;
	qsort(disk->fsm, disk->num_fsm_entries, sizeof(struct fsm_entry), compar_fsm);
	cur = 0;
	num_fsm_entries = disk->num_fsm_entries;
	next_sector = disk->fsm[0].lsn + disk->fsm[0].size;
	for (i = 1; i < num_fsm_entries; i++) {
		/* FSM entries shouldn't overlap */
		if (next_sector > disk->fsm[i].lsn) {
			LOG_WARN("*** coalesce_fsm(): Free space entries overlap.  Fixing.\n");
		}
		if (next_sector >= disk->fsm[i].lsn) {
			LOG_DEBUG("coalesce_fsm(): Merging %06X(%d) + %06X(%d)", disk->fsm[cur].lsn, disk->fsm[cur].size, disk->fsm[i].lsn, disk->fsm[i].size);
			disk->fsm[cur].size += disk->fsm[i].size;
			disk->fsm[cur].size -= (next_sector - disk->fsm[i].lsn);
			disk->num_fsm_entries--;
			disk->modified = 1;
			LOG_DEBUG(" = %06X(%d)\n", disk->fsm[cur].lsn, disk->fsm[cur].size);
		} else {
			cur++;
			disk->fsm[cur].lsn = disk->fsm[i].lsn;
			disk->fsm[cur].size = disk->fsm[i].size;
		}
		next_sector = disk->fsm[cur].lsn + disk->fsm[cur].size;
	}
}

int rewrite_fsm(struct ADFS_disk *disk) {
	struct timeval tp;
	int i;
	int free_sectors = 0;
	if (disk == NULL)
		return 0;
	/* Set disk id to something random */
	gettimeofday(&tp, NULL);
	disk->disk_id = tp.tv_sec & 0xffff;
	memset(gbuf, 0, SECTOR_SIZE * 2);
	coalesce_fsm(disk);
	for (i = 0; i < disk->num_fsm_entries; i++) {
		gbuf[i*3] = disk->fsm[i].lsn & 0xff;
		gbuf[i*3+1] = (disk->fsm[i].lsn >> 8) & 0xff;
		gbuf[i*3+2] = (disk->fsm[i].lsn >> 16) & 0xff;
		gbuf[i*3+0x100] = disk->fsm[i].size & 0xff;
		gbuf[i*3+0x101] = (disk->fsm[i].size >> 8) & 0xff;
		gbuf[i*3+0x102] = (disk->fsm[i].size >> 16) & 0xff;
		free_sectors += disk->fsm[i].size;
	}
	gbuf[0xfc] = disk->total_sectors & 0xff;
	gbuf[0xfd] = (disk->total_sectors >> 8) & 0xff;
	gbuf[0xfe] = (disk->total_sectors >> 16) & 0xff;
	gbuf[0x1fb] = disk->disk_id & 0xff;
	gbuf[0x1fc] = (disk->disk_id >> 8) & 0xff;
	gbuf[0x1fd] = disk->boot_option;
	gbuf[0x1fe] = disk->num_fsm_entries * 3;

	gbuf[SECTOR_SIZE - 1] = compute_checksum(gbuf);
	gbuf[(2*SECTOR_SIZE) - 1] = compute_checksum(gbuf + SECTOR_SIZE);
	LOG_DEBUG("rewrite_fsm(): free space now = %d\n", free_sectors);
	if (!write_sector(disk, 0, gbuf))
		return 0;
	if (!write_sector(disk, 1, gbuf + SECTOR_SIZE))
		return 0;
	return 1;
}
