/*  libadfs - library to manipulate old BBC Micro/Master ADFS images
 *  Copyright (C) 2008  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  */

#include "adfs.h"
#include "debug.h"
#include "disk.h"
#include "file.h"
#include "dir.h"
#include "dirent.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

/* ADFS file operations */

/* Called by read and write: ensures sector cache corresponds to file offset,
 * writing out old cache if different. */
static int update_sector_cache(ADFS_FILE *fd) {
	int new_lsn;
	if (fd == NULL) {
		errno = EBADF;
		return -1;
	}
	/* ADFS is simple like this - other filesystems would need a more
	 * complex offset to sector mapping. */
	new_lsn = fd->dirent->start_lsn + (fd->offset / SECTOR_SIZE);
	if (new_lsn == fd->cached_lsn)
		return 0;
	if (fd->cached_lsn != -1 && fd->sec_modified && (fd->oflag & O_ACCMODE) != O_RDONLY) {
		write_sector(fd->dirent->disk, fd->cached_lsn, fd->sec_data);
	}
	fd->cached_lsn = new_lsn;
	read_sector(fd->dirent->disk, fd->cached_lsn, fd->sec_data);
	fd->sec_modified = 0;
	return 0;
}

/* The basic file open call.  Uses ADFS_open_dirent() to open existing file,
 * or creates a new one if O_CREAT is specified. */
ADFS_FILE *ADFS_open(struct ADFS_disk *disk, const char *path,
		int oflag, int mode) {
	ADFS_DIR *dir;
	struct ADFS_dirent *ent;
	ADFS_FILE *fd;

	if (find_dir_and_ent(disk, path, &dir, &ent) < 0)
		return NULL;

	if (dir == NULL && ent) {
		/* path refers to / or . */
		return ADFS_open_dirent(ent, oflag, mode);
	}
	if (dir == NULL) {
		errno = ENOTDIR;
		return NULL;
	}
	/* parent_dir_index == -2 implies entry didn't exist, but a dummy
	 * entry with known data has been supplied in case we want to create
	 * it. */
	if (ent && ent->parent_dir_index >= -1) {
		ADFS_closedir(dir);
		/* Want CREAT|EXCL? */
		if ((oflag & (O_CREAT|O_EXCL)) == (O_CREAT|O_EXCL)) {
			errno = EEXIST;
			return NULL;
		}
		/* Otherwise it's all good */
		fd = ADFS_open_dirent(ent, oflag, ent->attr);
		return fd;
	}

	/* Entry doesn't exist */
	if (!(oflag & O_CREAT)) {
		ADFS_closedir(dir);
		errno = ENOENT;
		return NULL;
	}

	/* Want to create it */
	ent->attr = mode;

	if (add_to_directory(dir, ent) < 0) {
		ADFS_closedir(dir);
		return NULL;
	}
	ADFS_closedir(dir);

	return ADFS_open_dirent(ent, oflag, mode);
}

/* Take a dirent and wrap it in an ADFS_FILE * */
ADFS_FILE *ADFS_open_dirent(struct ADFS_dirent *dirent, int oflag, int mode) {
	ADFS_FILE *fd = malloc(sizeof(ADFS_FILE));
	(void)mode;
	if (fd == NULL)
		return NULL;
	LOG_DEBUG("ADFS_open_dirent(): name=%s lsn=%06X parent_lsn=%06X parent_index=%d\n", dirent->name, dirent->start_lsn, dirent->parent_dir_lsn, dirent->parent_dir_index);
	fd->oflag = oflag & O_ACCMODE;
	fd->dirent = dup_dirent(dirent);
	fd->cached_lsn = -1;
	if (oflag & O_APPEND) {
		ADFS_lseek(fd, 0, SEEK_END);
	} else {
		ADFS_lseek(fd, 0, SEEK_SET);
	}
	if (oflag & O_TRUNC)
		ADFS_ftruncate(fd, 0);
	update_sector_cache(fd);
	return fd;
}

ADFS_FILE *ADFS_creat(struct ADFS_disk *disk, const char *path, int mode) {
	return ADFS_open(disk, path, O_CREAT | O_TRUNC | O_WRONLY, mode);
}


int ADFS_close(ADFS_FILE *fd) {
	struct ADFS_dirent parent_ent;
	ADFS_DIR *parent_dir;
	if (fd == NULL) {
		errno = EBADF;
		return -1;
	}
	LOG_DEBUG("ADFS_close(): %s lsn=%06X parent_lsn=%06X parent_index=%d\n", fd->dirent->name, fd->dirent->start_lsn, fd->dirent->parent_dir_lsn, fd->dirent->parent_dir_index);
	/* Flush cached sector */
	ADFS_fsync(fd);
	/* Update file metadata if changed and associated with valid parent
	 * directory and does not have itself as parent (root directory) */
	if ((fd->oflag & O_ACCMODE) != O_RDONLY
			&& fd->dirent->start_lsn != fd->dirent->parent_dir_lsn
			&& fd->dirent->parent_dir_lsn >= 2
			&& fd->dirent->modified) {
		parent_dirent(&parent_ent, fd->dirent);
		parent_dir = ADFS_opendir_dirent(&parent_ent);
		if (parent_dir == NULL) {
			LOG_DEBUG("ADFS_close(): Couldn't open parent directory at %06X\n", parent_ent.start_lsn);
		} else {
			update_directory_entry(parent_dir, fd->dirent);
			/* Don't update directory's parent when it closes */
			parent_dir->fd->dirent->modified = 0;
			ADFS_closedir(parent_dir);
		}
	}
	free(fd->dirent);
	free(fd);
	return 0;
}

/* Just flush the currently cached sector. */
int ADFS_fsync(ADFS_FILE *fd) {
	if (fd == NULL)
		return -1;
	if (fd->cached_lsn != -1 && fd->sec_modified && (fd->oflag & O_ACCMODE) != O_RDONLY) {
		write_sector(fd->dirent->disk, fd->cached_lsn, fd->sec_data);
	}
	return 0;
}

/* Proper POSIX has a separate struct for stat buffers, but it's only
 * going to contain the same information... */
int ADFS_fstat(ADFS_FILE *fd, struct ADFS_dirent *buf) {
	if (fd == NULL) {
		errno = EBADF;
		return -1;
	}
	if (buf == NULL) {
		errno = EFAULT;
		return -1;
	}
	memcpy(buf, fd->dirent, sizeof(struct ADFS_dirent));
	return 0;
}

/* Known attributes are 'load=NUMBER', 'exec=NUMBER', 'mode=STRING' and
 * 'locked' */
int ADFS_set_attribute(ADFS_FILE *fd, const char *attr, const char *value) {
	if (fd == NULL) {
		errno = EBADF;
		return -1;
	}
	if (0 == strcmp(attr, "load")) {
		fd->dirent->load_address = strtoll(value, NULL, 0);
		fd->dirent->modified = 1;
	} else if (0 == strcmp(attr, "exec")) {
		fd->dirent->exec_address = strtoll(value, NULL, 0);
		fd->dirent->modified = 1;
	} else if (0 == strcmp(attr, "mode")) {
		fd->dirent->attr &= ~ADFS_ATTR_DIR;
		if (strchr(value, 'r'))
			fd->dirent->attr |= ADFS_ATTR_READ;
		if (strchr(value, 'w'))
			fd->dirent->attr |= ADFS_ATTR_WRITE;
		if (strchr(value, 'l') || strchr(value, 'p'))
			fd->dirent->attr |= ADFS_ATTR_LOCKED;
		fd->dirent->modified = 1;
	} else if (0 == strcmp(attr, "locked")) {
		fd->dirent->attr |= ADFS_ATTR_LOCKED;
		fd->dirent->modified = 1;
	}
	return 0;
}

int ADFS_ftruncate(ADFS_FILE *fd, off_t length) {
	if (fd == NULL) {
		errno = EBADF;
		return -1;
	}
	if ((fd->oflag & O_ACCMODE) == O_RDONLY) {
		errno = EBADF;
		return -1;
	}
	return truncate_dirent(fd->dirent, length);
}

off_t ADFS_lseek(ADFS_FILE *fd, off_t offset, int whence) {
	if (fd == NULL) {
		errno = EBADF;
		return -1;
	}
	switch (whence) {
		case SEEK_SET:
			fd->offset = offset;
			break;
		case SEEK_CUR:
			fd->offset = fd->offset + offset;
			break;
		case SEEK_END:
			fd->offset = fd->dirent->length + offset;
			break;
		default:
			errno = EINVAL;
			return -1;
	}
	return fd->offset;
}

/* read() and write() do their stuff in chunks of length SECTOR_SIZE */
ssize_t ADFS_read(ADFS_FILE *fd, void *buf, size_t nbytes) {
	uint8_t *bufp = buf;
	int rbytes = 0;
	if (fd == NULL || (fd->oflag & O_ACCMODE) == O_WRONLY) {
		errno = EBADF;
		return -1;
	}
	if (bufp == NULL) {
		errno = EFAULT;
		return -1;
	}
	if (fd->offset < 0 || (int)(fd->offset + nbytes) < 0) {
		errno = EINVAL;
		return -1;
	}
	while (nbytes > 0) {
		int sec_offset = fd->offset % SECTOR_SIZE;
		unsigned int rsize;
		if (sec_offset + nbytes >= SECTOR_SIZE) {
			rsize = SECTOR_SIZE - sec_offset;
		} else {
			rsize = nbytes;
		}
		if (fd->offset + rsize > fd->dirent->length) {
			rsize = fd->dirent->length - fd->offset;
		}
		if (rsize <= 0)
			return rbytes;
		if (update_sector_cache(fd) < 0) {
			LOG_DEBUG("ADFS_read(): update_sector_cache() failed\n");
			return -1;
		}
		memcpy(bufp, fd->sec_data + sec_offset, rsize);
		bufp += rsize;
		rbytes += rsize;
		fd->offset += rsize;
		nbytes -= rsize;
	}
	return rbytes;
}

ssize_t ADFS_write(ADFS_FILE *fd, const void *buf, size_t nbytes) {
	const uint8_t *bufp = buf;
	int wbytes = 0;
	if (fd == NULL || (fd->oflag & O_ACCMODE) == O_RDONLY) {
		errno = EBADF;
		return -1;
	}
	if (bufp == NULL) {
		errno = EFAULT;
		return -1;
	}
	if (fd->offset < 0 || (int)(fd->offset + nbytes) < 0) {
		errno = EINVAL;
		return -1;
	}
	while (nbytes > 0) {
		int sec_offset = fd->offset % SECTOR_SIZE;
		unsigned int wsize;
		if (sec_offset + nbytes >= SECTOR_SIZE) {
			wsize = SECTOR_SIZE - sec_offset;
		} else {
			wsize = nbytes;
		}
		if (fd->offset + wsize > fd->dirent->length) {
			if (ADFS_ftruncate(fd, fd->offset + wsize) < 0) {
				if (wbytes > 0)
					return wbytes;
				return -1;
			}
		}
		if (update_sector_cache(fd) < 0) {
			LOG_DEBUG("ADFS_write(): update_sector_cache() failed\n");
			return -1;
		}
		memcpy(fd->sec_data + sec_offset, bufp, wsize);
		fd->sec_modified = 1;
		bufp += wsize;
		wbytes += wsize;
		fd->offset += wsize;
		nbytes -= wsize;
	}
	return wbytes;
}

int ADFS_remove(struct ADFS_disk *disk, const char *path) {
	ADFS_DIR *dir, *tmp_dir;
	struct ADFS_dirent *ent, *tmp;
	int ret;

	if (find_dir_and_ent(disk, path, &dir, &ent) < 0)
		return -1;
	if (dir == NULL || ent == NULL) {
		errno = ENOENT;
		return -1;
	}

	LOG_DEBUG("ADFS_remove(): dir start=%06X ent start=%06X\n", dir->fd->dirent->start_lsn, ent->start_lsn);

	/* If ent refers to a directory and it opens as a valid directory,
	 * check it has no entries before removing. */
	if (ent->attr & ADFS_ATTR_DIR) {
		tmp_dir = ADFS_opendir_dirent(ent);
		if (tmp_dir) {
			tmp_dir->index = 0;  /* skip . and .. */
			tmp = ADFS_readdir(tmp_dir);
			if (tmp) {
				LOG_DEBUG("ADFS_remove(): directory at %06X not empty: contains '%s'\n", dir->fd->dirent->start_lsn, tmp->name);
				ADFS_closedir(tmp_dir);
				errno = ENOTEMPTY;
				return -1;
			}
			ADFS_closedir(tmp_dir);
		}
	}

	if (truncate_dirent(ent, 0) < 0)
		return -1;

	ret = remove_from_directory(dir, ent->parent_dir_index);
	ADFS_closedir(dir);
	return ret;
}

int ADFS_rename(struct ADFS_disk *disk, const char *old, const char *new) {
	ADFS_DIR *from_dir, *to_dir;
	struct ADFS_dirent *ent, from_ent, to_ent;

	/* Find entry for old name */
	if (find_dir_and_ent(disk, old, &from_dir, &ent) < 0)
		return -1;
	if (from_dir == NULL || ent == NULL || ent->parent_dir_index < 0) {
		errno = ENOENT;
		return -1;
	}
	copy_dirent(&from_ent, ent);

	/* Find directory for new name, ensure entry for it doesn't exist */
	if (find_dir_and_ent(disk, new, &to_dir, &ent) < 0)
		return -1;
	if (to_dir == NULL || ent == NULL) {
		errno = EFAULT;
		return -1;
	}
	if (ent->parent_dir_index >= 0) {
		errno = EEXIST;
		return -1;
	}
	copy_dirent(&to_ent, ent);

	/* Fill in remaining details for new entry */
	to_ent.attr = from_ent.attr;
	to_ent.load_address = from_ent.load_address;
	to_ent.exec_address = from_ent.exec_address;
	to_ent.length = from_ent.length;
	to_ent.start_lsn = from_ent.start_lsn;
	to_ent.cycle_number = from_ent.cycle_number;

	/* Remove old entry, add new entry */
	remove_from_directory(from_dir, from_ent.parent_dir_index);
	ADFS_closedir(from_dir);
	add_to_directory(to_dir, &to_ent);
	ADFS_closedir(to_dir);
	return 0;
}
