/*  fstool - manipulate old disk formats
 *  Copyright (C) 2008  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *  */

/* Currently ADFS only, using libadfs */

#include "config.h"
#include "portalib.h"
#include "adfs/adfs.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <inttypes.h>
#include <libgen.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_OPTRESET
# define RESET_GETOPT do { optreset = optind = 1; } while (0)
#else
# define RESET_GETOPT do { optind = 1; } while (0)
#endif

static int cmd_ls(int argc, char **argv);
static int cmd_cat(int argc, char **argv);
static int cmd_mkdir(int argc, char **argv);
static int cmd_attr(int argc, char **argv);
static int cmd_put(int argc, char **argv);
static int cmd_get(int argc, char **argv);
static int cmd_rm(int argc, char **argv);
static int cmd_mv(int argc, char **argv);

struct {
	const char *command;
	int (*dispatch)(int, char **);
	const char *desc;
} command_list[] = {
	{ "ls", cmd_ls, "show directory listing" },
	{ "dir", cmd_ls, "\"" },
	{ "cat", cmd_cat, "dump file contents to stdout" },
	{ "type", cmd_cat, "\"" },
	{ "mkdir", cmd_mkdir, "create directory" },
	{ "attr", cmd_attr, "print file attributes" },
	{ "put", cmd_put, "copy file to device" },
	{ "get", cmd_get, "copy file from device" },
	{ "rm", cmd_rm, "remove file" },
	{ "del", cmd_rm, "\"" },
	{ "mv", cmd_mv, "rename file" },
	{ "rename", cmd_mv, "\"" },
	{ NULL, NULL, NULL }
};

static struct ADFS_disk *disk;

static void main_usage(const char *prog) {
	int i;
	fprintf(stderr, "usage: %s [OPTION] cmd [OPTION]...\n\n", prog);
	fprintf(stderr, "  -t TYPE   specify fs type [adfs]\n");
	fprintf(stderr, "  -d FILE   open FILE as disk image [scsi0.dat]\n");
	fprintf(stderr, "  -C DIR    chdir to DIR before doing anything\n");
	fprintf(stderr, "  -h        show this help and exit\n\n");
	fprintf(stderr, "cmd is one of:\n");
	for (i = 0; command_list[i].command; i++) {
		fprintf(stderr, "%8s   %s\n", command_list[i].command, command_list[i].desc);
	}
}

int main(int argc, char **argv) {
	char *cmd = argv[0];
	char *fs_type = strdup("adfs");
	char *filename = strdup("scsi0.dat");
	char *start_dir = strdup("/");
	int opt, i;

	while ((opt = getopt(argc, argv, "t:d:C:h?")) != -1) {
		switch (opt) {
			case 't':
				/* not used yet */
				free(fs_type);
				fs_type = strdup(optarg);
				break;
			case 'd':
				free(filename);
				filename = strdup(optarg);
				break;
			case 'C':
				free(start_dir);
				start_dir = strdup(optarg);
				break;
			case 'h': case '?':
				main_usage(cmd);
				return 0;
			default:
				main_usage(cmd);
				exit(1);
		}
	}
	argc -= optind;
	argv += optind;
	RESET_GETOPT;

	if (argc < 1) {
		main_usage(cmd);
		exit(1);
	}

	disk = ADFS_open_disk(filename, O_RDWR);
	if (disk == NULL) {
		perror(filename);
		exit(1);
	}

	ADFS_chdir(disk, start_dir);
	free(start_dir);

	for (i = 0; command_list[i].command; i++) {
		if (0 == strcmp(command_list[i].command, argv[0])) {
			i = command_list[i].dispatch(argc, argv);
			ADFS_close_disk(disk);
			return i;
		}
	}
	fprintf(stderr, "%s: command not known\n", argv[0]);
	exit(1);
}

/**********************************************************************/

struct attribute_list {
	struct attribute_list *next;
	char *attr;
	char *value;
};

static void add_attribute(struct attribute_list **list, const char *str) {
	struct attribute_list *new;
	if (list == NULL) return;
	new = malloc(sizeof(struct attribute_list));
	if (new == NULL) return;
	new->next = NULL;
	new->value = strdup(str);
	new->attr = strsep(&new->value, "=");
	for ( ; *list; list = &(*list)->next);
	*list = new;
}

static void free_attribute_list(struct attribute_list *list) {
	struct attribute_list *tmp;
	while (list) {
		tmp = list->next;
		free(list->attr);
		free(list);
		list = tmp;
	}
}

static struct attribute_list *ADFS_attribute_list(ADFS_FILE *fd) {
	struct ADFS_dirent stat_buf;
	struct attribute_list *list = NULL;
	char buf[32];
	if (ADFS_fstat(fd, &stat_buf) < 0)
		return NULL;
	if (stat_buf.attr & ADFS_ATTR_DIR)
		add_attribute(&list, "dir");
	if (stat_buf.attr & ADFS_ATTR_LOCKED)
		add_attribute(&list, "locked");
	sprintf(buf, "load=0x%X", stat_buf.load_address);
	add_attribute(&list, buf);
	sprintf(buf, "exec=0x%X", stat_buf.exec_address);
	add_attribute(&list, buf);
	sprintf(buf, "length=0x%X", stat_buf.length);
	add_attribute(&list, buf);
	return list;
}

/**********************************************************************/

static void cmd_ls_usage(char *prog) {
	fprintf(stderr, "usage: %s [OPTION] [PATH]\n\n", prog);
	fprintf(stderr, "  -a   list all files\n");
	fprintf(stderr, "  -l   long listing\n");
	fprintf(stderr, "  -x   fs-specific long listing\n");
	fprintf(stderr, "  -h   show this help and exit\n");
}

static int cmd_ls(int argc, char **argv) {
	ADFS_DIR *dir;
	struct ADFS_dirent *ent;
	char *path;
	int long_listing = 0;
	int extended_listing = 0;
	int all_files = 0;
	int opt;

	while ((opt = getopt(argc, argv, "alxh?")) != -1) {
		switch (opt) {
			case 'a':
				all_files = 1;
				break;
			case 'l':
				long_listing = 1;
				break;
			case 'x':
				extended_listing = 1;
				break;
			case 'h': case '?':
				cmd_ls_usage(argv[0]);
				return 0;
			default:
				cmd_ls_usage(argv[0]);
				exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc >= 1) {
		path = strdup(argv[0]);
	} else {
		path = strdup(".");
	}

	if (extended_listing) {
		int ret = ADFS_extended_ls(disk, path);
		if (ret < 0)
			perror(path);
		free(path);
		return ret;
	}

	if ((dir = ADFS_opendir(disk, path)) == NULL) {
		perror(path);
		free(path);
		return 1;
	}
	while ((ent = ADFS_readdir(dir))) {
		if (!all_files && ent->name[0] == '.')
			continue;
		if (long_listing) {
			char isdir = (ent->attr & ADFS_ATTR_DIR) ? 'd' : '-';
			char canread = (ent->attr & ADFS_ATTR_READ) ? 'r' : '-';
			char canwrite = (ent->attr & ADFS_ATTR_WRITE) ? 'w' : '-';
			char canexec = (ent->attr & ADFS_ATTR_EXEC) ? 'x' : '-';
			printf("%c%c%c%c %7d %s\n", isdir, canread, canwrite, canexec, ent->length, ent->name);
		} else {
			printf("%s\n", ent->name);
		}
	}
	ADFS_closedir(dir);
	free(path);
	return 0;
}

static int cmd_cat(int argc, char **argv) {
	char *filename;
	ADFS_FILE *file;
	int rbytes;
	char buf[256];
	int i;

	if (argc < 2) {
		fprintf(stderr, "usage: %s file...\n", argv[0]);
		exit(1);
	}
	for (i = 1; i < argc; i++) {
		filename = strdup(argv[1]);
		file = ADFS_open(disk, filename, O_RDONLY, 0);
		if (file == NULL) {
			perror(filename);
		} else {
			while ((rbytes = ADFS_read(file, buf, sizeof(buf))) > 0) {
				fwrite(buf, rbytes, 1, stdout);
			}
			ADFS_close(file);
		}
		free(filename);
	}
	return 0;
}

static int cmd_mkdir(int argc, char **argv) {
	char *filename;
	int ret;

	if (argc < 2) {
		fprintf(stderr, "usage: %s file\n", argv[0]);
		exit(1);
	}
	filename = strdup(argv[1]);
	ret = ADFS_mkdir(disk, filename);
	if (ret < 0)
		perror(filename);
	free(filename);
	return ret;
}

static void cmd_put_usage(char *prog) {
	fprintf(stderr, "usage: %s [OPTION] file [dest]\n\n", prog);
	fprintf(stderr, "  -a ATTR=VALUE   set attribute on file\n");
	fprintf(stderr, "  -h              show this help and exit\n");
}

static int cmd_put(int argc, char **argv) {
	char *cmd = argv[0];
	char *src, *dest = NULL;
	int srcfd;
	ADFS_FILE *destfd;
	struct stat statbuf;
	char buf[256];
	int bytes;
	int opt;
	struct attribute_list *attr_list = NULL, *attr;

	while ((opt = getopt(argc, argv, "a:h?")) != -1) {
		switch (opt) {
			case 'a':
				add_attribute(&attr_list, optarg);
				break;
			case 'h': case '?':
				cmd_put_usage(cmd);
				free_attribute_list(attr_list);
				return 0;
			default:
				cmd_put_usage(cmd);
				free_attribute_list(attr_list);
				return 1;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc < 1) {
		cmd_put_usage(cmd);
		free_attribute_list(attr_list);
		return 1;
	}
	src = strdup(argv[0]);
	if (argc >= 2)
		dest = argv[1];
	if (dest == NULL)
		dest = basename(argv[0]);

	if (stat(src, &statbuf) < 0) {
		perror(src);
		free_attribute_list(attr_list);
		return 1;
	}
	srcfd = open(src, O_RDONLY);
	if (srcfd < -1) {
		perror(src);
		free(src);
		free_attribute_list(attr_list);
		return 1;
	}
	free(src);

	destfd = ADFS_open(disk, dest, O_CREAT | O_TRUNC | O_WRONLY, ADFS_ATTR_READ | ADFS_ATTR_WRITE);
	if (destfd == NULL) {
		perror(dest);
		free_attribute_list(attr_list);
		return 1;
	}
	for (attr = attr_list; attr; attr = attr->next) {
		ADFS_set_attribute(destfd, attr->attr, attr->value);
	}
	if (ADFS_ftruncate(destfd, statbuf.st_size) < 0) {
		perror(dest);
		free_attribute_list(attr_list);
		return 1;
	}
	while ((bytes = read(srcfd, buf, sizeof(buf))) > 0) {
		ADFS_write(destfd, buf, bytes);
	}
	ADFS_close(destfd);
	close(srcfd);
	free_attribute_list(attr_list);
	return 0;
}

static int cmd_attr(int argc, char **argv) {
	char *cmd = argv[0];
	ADFS_FILE *fd;
	struct attribute_list *attr_list, *attr;

	if (argc < 2) {
		fprintf(stderr, "usage: %s file\n", cmd); 
		return 1;
	}
	fd = ADFS_open(disk, argv[1], O_RDONLY, 0);
	if (fd == NULL) {
		perror(argv[1]);
		return 1;
	}
	attr_list = ADFS_attribute_list(fd);
	for (attr = attr_list; attr; attr = attr->next) {
		if (attr->value) {
			printf("%s=%s\n", attr->attr, attr->value);
		} else {
			printf("%s\n", attr->attr);
		}
	}
	free_attribute_list(attr_list);
	return 0;
}

static int cmd_get(int argc, char **argv) {
	char *src, *dest, *destc;
	ADFS_FILE *srcfd;
	int destfd;
	struct ADFS_dirent ent_buf;
	uint8_t buf[256];
	int bytes;

	if (argc < 2) {
		fprintf(stderr, "usage: %s src [dest]\n", argv[0]); 
		return 1;
	}
	if ((src = strdup(argv[1])) == NULL)
		return 1;

	srcfd = ADFS_open(disk, src, O_RDONLY, 0);
	if (srcfd == NULL) {
		perror(src);
		free(src);
		return 1;
	}
	free(src);

	if (argc > 2) {
		destc = strdup(argv[2]);
		dest = destc;
	} else {
		destc = strdup(argv[1]);
		dest = basename(destc);
	}
	if (destc == NULL)
		return 1;

	ADFS_fstat(srcfd, &ent_buf);
	if (ent_buf.attr & ADFS_ATTR_DIR) {
		mkdir(dest, S_IRUSR | S_IWUSR | S_IXUSR);
		ADFS_close(srcfd);
		free(destc);
		return 0;
	}
	destfd = open(dest, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
	if (destfd < 0) {
		perror(dest);
		free(destc);
		return 1;
	}
	free(destc);
	while ((bytes = ADFS_read(srcfd, buf, sizeof(buf))) > 0) {
		write(destfd, buf, bytes);
	}
	close(destfd);

	ADFS_close(srcfd);
	return 0;
}

static int cmd_rm(int argc, char **argv) {
	char *filename;
	int ret;

	if (argc < 2) {
		fprintf(stderr, "usage: %s file\n", argv[0]);
		return 1;
	}

	if ((filename = strdup(argv[1])) == NULL)
		return 1;
	ret = ADFS_remove(disk, filename);
	if (ret < 0)
		perror(filename);
	free(filename);
	return ret;
}

static int cmd_mv(int argc, char **argv) {
	char *old, *new;
	int ret;

	if (argc < 3) {
		fprintf(stderr, "usage: %s old-name new-name\n", argv[0]);
		return 1;
	}
	if ((old = strdup(argv[1])) == NULL)
		return 1;
	if ((new = strdup(argv[2])) == NULL) {
		free(old);
		return 1;
	}

	ret = ADFS_rename(disk, old, new);
	if (ret < 0)
		perror(NULL);
	free(old);
	free(new);
	return ret;
}
